import json
import logging
import random
import string
import time
from collections.abc import Sequence
from datetime import datetime
from http import HTTPStatus
from io import StringIO
from ssl import SSLContext
from typing import Any, ClassVar, Final
from urllib.parse import quote

from aiohttp import ClientResponse, ClientSession, ClientTimeout, TCPConnector
from awsauth import S3Auth
from requests import Request

from .exceptions import (
    RGWAdminException, AccessDenied, UserExists,
    InvalidAccessKey, InvalidSecretKey, InvalidKeyType,
    KeyExists, EmailExists, SubuserExists, InvalidAccess,
    IndexRepairFailed, BucketNotEmpty, ObjectRemovalFailed,
    BucketUnlinkFailed, BucketLinkFailed, NoSuchObject,
    IncompleteBody, InvalidCap, NoSuchCap,
    InternalError, NoSuchUser, NoSuchBucket, NoSuchKey,
    ServerDown, InvalidQuotaType, InvalidArgument, BucketAlreadyExists,
)

log: Final = logging.getLogger(__name__)
LETTERS: Final = string.ascii_letters


class RGWAdmin:
    _access_key: str
    _secret_key: str
    _server: str
    _admin: str
    _response: str
    _ssl_context: SSLContext | None
    _verify: bool
    _protocol: str
    _timeout: float | None
    _session: ClientSession | None

    connection: ClassVar['RGWAdmin']

    metadata_types: ClassVar[list[str]] = ['user', 'bucket', 'bucket.instance']

    def __init__(
        self,
        access_key: str,
        secret_key: str,
        server: str,
        admin: str = 'admin',
        response: str = 'json',
        ssl_context: SSLContext | None = None,
        secure: bool = True,
        verify: bool = True,
        timeout: float | None = None,
        pool_connections: bool = False,
    ) -> None:
        self._access_key = access_key
        self._secret_key = secret_key
        self._server = server
        self._admin = admin
        self._response = response
        self._session = None

        # ssl support
        self._ssl_context = ssl_context
        self._verify = verify
        if secure:
            self._protocol = 'https'
        else:
            self._protocol = 'http'

        self._timeout = timeout
        self._skip_auto_headers = ["Content-Type"]

        self._auth = S3Auth(self._access_key, self._secret_key, self._server)

        if pool_connections:
            self._session = self._get_session()

    async def __aenter__(self) -> 'RGWAdmin':
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.close()

    async def close(self) -> None:
        if self._session:
            await self._session.close()

    def _get_session(self) -> ClientSession:
        if self._ssl_context:
            ssl = self._ssl_context
        else:
            ssl = self._verify

        return ClientSession(
            connector=TCPConnector(ssl=ssl),
            skip_auto_headers=self._skip_auto_headers,
            timeout=ClientTimeout(self._timeout),
        )

    @classmethod
    def connect(cls, **kwargs: Any) -> None:
        """Establish a new connection to RGWAdmin

        Only one connection can be active in any single process
        """
        cls.set_connection(cls(**kwargs))

    @classmethod
    def set_connection(cls, connection: 'RGWAdmin') -> None:
        """Set a connection for the RGWAdmin session to use."""
        cls.connection = connection

    @classmethod
    def get_connection(cls) -> 'RGWAdmin':
        """Return the RGWAdmin connection that was set"""
        return cls.connection

    def __repr__(self) -> str:
        return "%s (%s)" % (self.__class__.__name__, self.get_base_url())

    def __str__(self) -> str:
        returning = self.__repr__()
        returning += '\nAccess Key: %s\n' % self._access_key
        returning += 'Secret Key: ******\n'
        returning += 'Response Method: %s\n' % self._response
        if self._ssl_context is not None:
            returning += 'SSL Context: %s\n' % self._ssl_context
        return returning

    def get_base_url(self) -> str:
        '''Return a base URL.  I.e. https://ceph.server'''
        return '%s://%s' % (self._protocol, self._server)

    @staticmethod
    async def _load_request(r: ClientResponse) -> Any:
        '''Load the request given as JSON handling exceptions if necessary'''
        try:
            j = await r.json(content_type=None)
        except ValueError:
            # some calls in the admin API encode the info in the headers
            # instead of the body.  The code that follows is an ugly hack
            # due to the fact that there's a bug in the admin API we're
            # interfacing with.

            # set a default value for j in case we don't find json in the
            # headers below
            j = None

            # find a key with a '{', since this will hold the json response
            for k, v in r.headers.items():
                if '{' in k:
                    json_string = ":".join([k, v]).split('}')[0] + '}'
                    j = json.load(StringIO(json_string))
                    break

        if r.status == HTTPStatus.OK:
            return j
        elif r.status == HTTPStatus.NO_CONTENT:
            return None
        else:
            if j is not None:
                code = str(j.get('Code', 'InternalError'))
            else:
                raise ServerDown(r.status)

            for e in [AccessDenied, UserExists, InvalidAccessKey,
                      InvalidKeyType, InvalidSecretKey, KeyExists, EmailExists,
                      SubuserExists, InvalidAccess, InvalidArgument,
                      IndexRepairFailed, BucketNotEmpty, ObjectRemovalFailed,
                      BucketUnlinkFailed, BucketLinkFailed, NoSuchObject,
                      InvalidCap, NoSuchCap, NoSuchUser, NoSuchBucket,
                      NoSuchKey, IncompleteBody, BucketAlreadyExists,
                      InternalError]:
                if code == e.__name__:
                    raise e(j)

            raise RGWAdminException(code, raw=j)

    async def request(self, method: str, request: str, headers: dict | None = None, data: Any = None) -> Any:
        url = '%s%s' % (self.get_base_url(), request)
        log.debug('URL: %s' % url)
        log.debug('Access Key: %s' % self._access_key)
        log.debug('Verify: %s, SSL Context: %s' % (self._verify, self._ssl_context))

        # prepare headers for auth
        prepped = Request(method, url, headers=headers, auth=self._auth).prepare()
        prepped_headers = prepped.headers

        if data is not None:
            prepped_headers["Content-Length"] = str(len(data))

        request_params = {
            "method": method,
            "url": url,
            "headers": prepped_headers,
            "data": data,
        }

        if self._session:
            # use connection pool
            async with self._session.request(**request_params) as response:
                return await self._load_request(response)
        else:
            # do not use connection pool
            async with self._get_session() as session:
                async with session.request(**request_params) as response:
                    return await self._load_request(response)

    async def _request_metadata(
        self,
        method: str,
        metadata_type: str,
        params: dict | None = None,
        headers: dict | None = None,
        data: Any = None,
    ) -> Any:
        if metadata_type not in self.metadata_types:
            raise Exception("Bad metadata_type")

        if params is None:
            params = {}
        params = '&'.join(['%s=%s' % (k, v) for k, v in params.items()])
        request = '/%s/metadata/%s?%s' % (self._admin, metadata_type, params)
        return await self.request(
            method=method,
            request=request,
            headers=headers,
            data=data,
        )

    async def get_metadata(
        self,
        metadata_type: str,
        key: str | None = None,
        max_entries: int | None = None,
        marker: str | None = None,
        headers: dict | None = None,
    ) -> Any:
        ''' Returns a JSON object representation of the metadata '''
        params = {'format': self._response}
        if key is not None:
            params['key'] = key
        if marker is not None:
            params['marker'] = quote(marker)
        if max_entries is not None:
            params['max-entries'] = max_entries
        return await self._request_metadata(
            method='get',
            metadata_type=metadata_type,
            params=params,
            headers=headers,
        )

    async def put_metadata(self, metadata_type: str, key: str, json_string: str) -> Any:
        return await self._request_metadata(
            method='put',
            metadata_type=metadata_type,
            params={'key': key},
            headers={'Content-Type': 'application/json'},
            data=json_string,
        )

    # Alias for compatability:
    set_metadata = put_metadata

    async def delete_metadata(self, metadata_type: str, key: str) -> Any:
        return await self._request_metadata(
            method='delete',
            metadata_type=metadata_type,
            params={'key': key},
        )

    async def lock_metadata(self, metadata_type: str, key: str, lock_id: str, length: int) -> Any:
        params = {
            'lock': 'lock',
            'key': key,
            'lock_id': lock_id,
            'length': length,
        }
        return await self._request_metadata(
            method='post',
            metadata_type=metadata_type,
            params=params,
        )

    async def unlock_metadata(self, metadata_type: str, key: str, lock_id: str) -> Any:
        params = {
            'unlock': 'unlock',
            'key': key,
            'lock_id': lock_id,
        }
        return await self._request_metadata(
            method='post',
            metadata_type=metadata_type,
            params=params,
        )

    async def get_user(
        self,
        uid: str | None = None,
        access_key: str | None = None,
        stats: bool = False,
        sync: bool = False,
    ) -> Any:
        if uid is not None and access_key is not None:
            raise ValueError('Only one of uid and access_key is allowed')
        parameters = ''
        if uid is not None:
            parameters += '&uid=%s' % uid
        if access_key is not None:
            parameters += '&access-key=%s' % access_key
        parameters += '&stats=%s&sync=%s' % (stats, sync)
        return await self.request('get', '/%s/user?format=%s%s' % (self._admin, self._response, parameters))

    async def get_users(self) -> Any:
        return await self.get_metadata(metadata_type='user')

    async def create_user(
        self,
        uid: str,
        display_name: str,
        email: str | None = None,
        key_type: str = 's3',
        access_key: str | None = None,
        secret_key: str | None = None,
        user_caps: str | None = None,
        generate_key: bool = True,
        max_buckets: int | None = None,
        suspended: bool = False,
    ) -> Any:
        parameters = 'uid=%s&display-name=%s' % (uid, display_name)
        if email is not None:
            parameters += '&email=%s' % email
        if key_type is not None:
            parameters += '&key-type=%s' % key_type
        if access_key is not None:
            parameters += '&access-key=%s' % access_key
        if secret_key is not None:
            parameters += '&secret-key=%s' % secret_key
        if user_caps is not None:
            parameters += '&user-caps=%s' % user_caps
        parameters += '&generate-key=%s' % generate_key
        if max_buckets is not None:
            parameters += '&max-buckets=%s' % max_buckets
        parameters += '&suspended=%s' % suspended
        return await self.request('put', '/%s/user?format=%s&%s' % (self._admin, self._response, parameters))

    async def get_usage(
        self,
        uid: str | None = None,
        start: str | datetime | None = None,
        end: str | datetime | None = None,
        show_entries: bool = False,
        show_summary: bool = False,
    ) -> Any:
        parameters = ''
        if uid is not None:
            parameters += '&uid=%s' % uid
        if start is not None:
            parameters += '&start=%s' % start
        if end is not None:
            parameters += '&end=%s' % end
        parameters += '&show-entries=%s' % show_entries
        parameters += '&show-summary=%s' % show_summary
        return await self.request('get', '/%s/usage?format=%s%s' % (self._admin, self._response, parameters))

    async def trim_usage(
        self,
        uid: str | None = None,
        start: str | datetime | None = None,
        end: str | datetime | None = None,
        remove_all: bool = False,
    ) -> Any:
        parameters = ''
        if uid is not None:
            parameters += '&uid=%s' % uid
        if start is not None:
            parameters += '&start=%s' % start
        if end is not None:
            parameters += '&end=%s' % end
        parameters += '&remove-all=%s' % remove_all
        return await self.request('delete', '/%s/usage?format=%s%s' % (self._admin, self._response, parameters))

    async def modify_user(
        self,
        uid: str,
        display_name: str | None = None,
        email: str | None = None,
        key_type: str = 's3',
        access_key: str | None = None,
        secret_key: str | None = None,
        user_caps: str | None = None,
        generate_key: bool = False,
        max_buckets: int | None = None,
        suspended: bool | None = None,
    ) -> Any:
        parameters = 'uid=%s' % uid
        if display_name is not None:
            parameters += '&display-name=%s' % display_name
        if email is not None:
            parameters += '&email=%s' % email
        if key_type is not None:
            parameters += '&key-type=%s' % key_type
        if access_key is not None:
            parameters += '&access-key=%s' % access_key
        if secret_key is not None:
            parameters += '&secret-key=%s' % secret_key
        if user_caps is not None:
            parameters += '&user-caps=%s' % user_caps
        parameters += '&generate-key=%s' % generate_key
        if max_buckets is not None:
            parameters += '&max-buckets=%s' % max_buckets
        if suspended is not None:
            parameters += '&suspended=%s' % suspended
        return await self.request('post', '/%s/user?format=%s&%s' % (self._admin, self._response, parameters))

    async def get_quota(self, uid: str, quota_type: str) -> Any:
        if quota_type not in ['user', 'bucket']:
            raise InvalidQuotaType
        parameters = 'uid=%s&quota-type=%s' % (uid, quota_type)
        return await self.request('get', '/%s/user?quota&format=%s&%s' % (self._admin, self._response, parameters))

    async def get_user_quota(self, uid: str) -> Any:
        return await self.get_quota(uid=uid, quota_type='user')

    async def get_user_bucket_quota(self, uid: str) -> Any:
        '''Return the quota set on every bucket owned/created by a user'''
        return await self.get_quota(uid=uid, quota_type='bucket')

    @staticmethod
    def _quota(max_size_kb: int | None = None, max_objects: int | None = None, enabled: bool | None = None) -> str:
        quota = ''
        if max_size_kb is not None:
            quota += '&max-size-kb=%d' % max_size_kb
        if max_objects is not None:
            quota += '&max-objects=%d' % max_objects
        if enabled is not None:
            quota += '&enabled=%s' % str(enabled).lower()
        return quota

    async def set_user_quota(
        self,
        uid: str,
        quota_type: str,
        max_size_kb: int | None = None,
        max_objects: int | None = None,
        enabled: bool | None = None,
    ) -> Any:
        '''
        Set quotas on users and buckets owned by users

        If `quota_type` is user, then the quota applies to the user.  If
        `quota_type` is bucket, then the quota applies to buckets owned by
        the specified uid.

        If you want to set a quota on an individual bucket, then use
        set_bucket_quota() instead.
        '''
        if quota_type not in ['user', 'bucket']:
            raise InvalidQuotaType
        quota = self._quota(max_size_kb=max_size_kb, max_objects=max_objects, enabled=enabled)
        parameters = 'uid=%s&quota-type=%s%s' % (uid, quota_type, quota)
        return await self.request('put', '/%s/user?quota&format=%s&%s' % (self._admin, self._response, parameters))

    async def set_bucket_quota(
        self,
        uid: str,
        bucket: str,
        max_size_kb: int | None = None,
        max_objects: int | None = None,
        enabled: bool | None = None,
    ) -> Any:
        '''Set the quota on an individual bucket'''
        quota = self._quota(max_size_kb=max_size_kb, max_objects=max_objects, enabled=enabled)
        parameters = 'uid=%s&bucket=%s%s' % (uid, bucket, quota)
        return await self.request('put', '/%s/bucket?quota&format=%s&%s' % (self._admin, self._response, parameters))

    async def remove_user(self, uid: str, purge_data: bool = False) -> Any:
        parameters = 'uid=%s' % uid
        parameters += '&purge-data=%s' % purge_data
        return await self.request('delete', '/%s/user?format=%s&%s' % (self._admin, self._response, parameters))

    async def create_subuser(
        self,
        uid: str,
        subuser: str | None = None,
        secret_key: str | None = None,
        access_key: str | None = None,
        key_type: str | None = None,
        access: str | None = None,
        generate_secret: bool = False,
    ) -> Any:
        parameters = 'uid=%s' % uid
        if subuser is not None:
            parameters += '&subuser=%s' % subuser
        if secret_key is not None and access_key is not None:
            parameters += '&access-key=%s' % access_key
            parameters += '&secret-key=%s' % secret_key
        if key_type is not None and key_type.lower() in ['s3', 'swift']:
            parameters += '&key-type=%s' % key_type
        if access is not None:
            parameters += '&access=%s' % access
        parameters += '&generate-secret=%s' % generate_secret
        return await self.request('put', '/%s/user?subuser&format=%s&%s' % (self._admin, self._response, parameters))

    async def modify_subuser(
        self,
        uid: str,
        subuser: str,
        secret: str | None = None,
        key_type: str = 'swift',
        access: str | None = None,
        generate_secret: bool = False,
    ) -> Any:
        parameters = 'uid=%s&subuser=%s' % (uid, subuser)
        if secret is not None:
            parameters += '&secret=%s' % secret
        parameters += '&key-type=%s' % key_type
        if access is not None:
            parameters += '&access=%s' % access
        parameters += '&generate-secret=%s' % generate_secret
        return await self.request('post', '/%s/user?subuser&format=%s&%s' % (self._admin, self._response, parameters))

    async def remove_subuser(self, uid: str, subuser: str, purge_keys: bool = True) -> Any:
        parameters = 'uid=%s&subuser=%s&purge-keys=%s' % (uid, subuser, purge_keys)
        return await self.request('delete', '/%s/user?subuser&format=%s&%s' % (self._admin, self._response, parameters))

    async def create_key(
        self,
        uid: str,
        subuser: str | None = None,
        key_type: str = 's3',
        access_key: str | None = None,
        secret_key: str | None = None,
        generate_key: bool = True,
    ) -> Any:
        parameters = 'uid=%s' % uid
        if subuser is not None:
            parameters += '&subuser=%s' % subuser
        parameters += '&key-type=%s' % key_type
        if access_key is not None:
            parameters += '&access-key=%s' % access_key
        if secret_key is not None:
            parameters += '&secret-key=%s' % secret_key
        parameters += '&generate-key=%s' % generate_key
        return await self.request('put', '/%s/user?key&format=%s&%s' % (self._admin, self._response, parameters))

    async def remove_key(
        self,
        access_key: str,
        key_type: str | None = None,
        uid: str | None = None,
        subuser: str | None = None,
    ) -> Any:
        parameters = 'access-key=%s' % access_key
        if key_type is not None:
            parameters += '&key-type=%s' % key_type
        if uid is not None:
            parameters += '&uid=%s' % uid
        if subuser is not None:
            parameters += '&subuser=%s' % subuser
        return await self.request('delete', '/%s/user?key&format=%s&%s' % (self._admin, self._response, parameters))

    async def get_buckets(self) -> Any:
        '''Returns a list of all buckets in the radosgw'''
        return await self.get_metadata(metadata_type='bucket')

    async def get_bucket(self, bucket: str | None = None, uid: str | None = None, stats: bool = False) -> Any:
        parameters = ''
        if bucket is not None:
            parameters += '&bucket=%s' % bucket
        if uid is not None:
            parameters += '&uid=%s' % uid
        parameters += '&stats=%s' % stats
        return await self.request('get', '/%s/bucket?format=%s%s' % (self._admin, self._response, parameters))

    async def check_bucket_index(self, bucket: str, check_objects: bool = False, fix: bool = False) -> Any:
        parameters = 'bucket=%s' % bucket
        parameters += '&check-objects=%s' % check_objects
        parameters += '&fix=%s' % fix
        return await self.request('get', '/%s/bucket?index&format=%s&%s' % (self._admin, self._response, parameters))

    async def remove_bucket(self, bucket: str, purge_objects: bool = False) -> Any:
        parameters = 'bucket=%s' % bucket
        parameters += '&purge-objects=%s' % purge_objects
        return await self.request('delete', '/%s/bucket?format=%s&%s' % (self._admin, self._response, parameters))

    async def unlink_bucket(self, bucket: str, uid: str) -> Any:
        parameters = 'bucket=%s&uid=%s' % (bucket, uid)
        return await self.request('post', '/%s/bucket?format=%s&%s' % (self._admin, self._response, parameters))

    async def link_bucket(self, bucket: str, bucket_id: str, uid: str) -> Any:
        # note that even though the Ceph docs say that bucket-id is optional
        # the API call will fail (InvalidArgument) if it is omitted.
        parameters = 'bucket=%s&bucket-id=%s&uid=%s' % (bucket, bucket_id, uid)
        return await self.request('put', '/%s/bucket?format=%s&%s' % (self._admin, self._response, parameters))

    async def remove_object(self, bucket: str, object_name: str) -> Any:
        parameters = 'bucket=%s&object=%s' % (bucket, object_name)
        return await self.request('delete', '/%s/bucket?object&format=%s&%s' % (self._admin, self._response, parameters))

    async def get_policy(self, bucket: str, object_name: str | None = None) -> Any:
        parameters = 'bucket=%s' % bucket
        if object_name is not None:
            parameters += '&object=%s' % object_name
        return await self.request('get', '/%s/bucket?policy&format=%s&%s' % (self._admin, self._response, parameters))

    async def add_capability(self, uid: str, user_caps: str) -> Any:
        parameters = 'uid=%s&user-caps=%s' % (uid, user_caps)
        return await self.request('put', '/%s/user?caps&format=%s&%s' % (self._admin, self._response, parameters))

    async def remove_capability(self, uid: str, user_caps: str) -> Any:
        parameters = 'uid=%s&user-caps=%s' % (uid, user_caps)
        return await self.request('delete', '/%s/user?caps&format=%s&%s' % (self._admin, self._response, parameters))

    async def get_bucket_instances(self) -> Any:
        '''Returns a list of all bucket instances in the radosgw'''
        return await self.get_metadata(metadata_type='bucket.instance')

    @staticmethod
    def parse_rados_datestring(s: str) -> time.struct_time:
        return time.strptime(s, "%Y-%m-%dT%H:%M:%S.%fZ")

    @staticmethod
    def gen_secret_key(size: int = 40, chars: Sequence = LETTERS + string.digits) -> str:
        return ''.join(random.choice(chars) for _ in range(size))



#!/usr/bin/env python

import logging
import unittest
import uuid
import random

import aiorgwadmin
from . import create_bucket, get_environment_creds

logging.basicConfig(level=logging.WARNING)


class RGWAdminTest(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.rgw = aiorgwadmin.RGWAdmin(pool_connections=True,
                                        **get_environment_creds())
        self.user1 = 'foo1209'
        self.user2 = 'foo1213'
        self.user3 = 'bar3142'
        self.secret = aiorgwadmin.RGWAdmin.gen_secret_key()
        user1 = await self.rgw.create_user(uid=self.user1,
                                           email='%s@example.com' % self.user1,
                                           display_name='Unit Test %s' % self.user1,
                                           secret_key=self.secret)
        user2 = await self.rgw.create_user(uid=self.user2,
                                           display_name='Unit Test %s' % self.user2,
                                           secret_key=self.secret)
        self.user1_obj = user1
        self.user2_obj = user2
        self.assertTrue(user1['user_id'] == self.user1)
        self.assertTrue(user2['user_id'] == self.user2)

    async def asyncTearDown(self):
        await self.rgw.remove_user(uid=self.user1, purge_data=True)
        await self.rgw.remove_user(uid=self.user2, purge_data=True)
        await self.rgw.close()

    async def test_modify_user(self):
        before_keys = self.user1_obj['keys']
        user = await self.rgw.modify_user(uid=self.user1,
                                          email='%s@test.com' % self.user1)
        self.assertTrue(user['email'] == '%s@test.com' % self.user1)

        # ensure that generate_key is False and that new keys are not
        # being created by default when modifying the user.
        self.assertEqual(before_keys, user['keys'])

        # generate_key=True should add an extra pair of keys
        user = await self.rgw.modify_user(uid=self.user1, generate_key=True)
        self.assertNotEqual(before_keys, user['keys'])
        self.assertEqual(len(before_keys) + 1, len(user['keys']))

        # modifying a user doesn't automatically un-suspend them
        # setup - suspend them first
        user = await self.rgw.modify_user(uid=self.user1, suspended=True)
        self.assertEqual(user['suspended'], 1)

        # test - modify their username without unsuspending them
        user = await self.rgw.modify_user(uid=self.user1,
                                          email='%s@test2.com' % self.user1)
        self.assertTrue(user['email'] == '%s@test2.com' % self.user1)
        self.assertEqual(user['suspended'], 1)

        # unsuspend the user to allow other tests to function
        user = await self.rgw.modify_user(uid=self.user1, suspended=False)
        self.assertEqual(user['suspended'], 0)

    async def test_duplicate_email(self):
        with self.assertRaises(aiorgwadmin.exceptions.EmailExists):
            await self.rgw.create_user(uid=self.user3,
                                       email='%s@example.com' % self.user1,
                                       display_name='Unit Test %s' % self.user3,
                                       secret_key=aiorgwadmin.RGWAdmin.gen_secret_key())

    async def test_get_user(self):
        user = await self.rgw.get_user(uid=self.user2)
        self.assertTrue(user['display_name'] == 'Unit Test %s' % self.user2)
        self.assertTrue('stats' not in user)

        user = await self.rgw.get_user(uid=self.user2, stats=True)
        stats = user['stats']
        self.assertEqual(stats['size'], 0)
        self.assertEqual(stats['size_actual'], 0)

    async def test_get_users(self):
        users = await self.rgw.get_users()
        self.assertTrue(self.user1 in users)
        self.assertTrue(self.user2 in users)

    async def test_user_quota(self):
        size = random.randint(1000, 1000000)
        await self.rgw.set_user_quota(uid=self.user1, quota_type='user',
                                      max_size_kb=size, enabled=True)
        user1_quota_info = await self.rgw.get_user_quota(uid=self.user1)
        self.assertTrue(size == user1_quota_info['max_size_kb'])

    async def test_bucket_quota(self):
        size = random.randint(1000, 1000000)
        bucket_name = f"bucket-{uuid.uuid4()}"
        await create_bucket(name=bucket_name, owner=self.user1)
        await self.rgw.set_bucket_quota(uid=self.user1, bucket=bucket_name,
                                  max_size_kb=size, enabled=True)
        bucket = await self.rgw.get_bucket(bucket=bucket_name)
        self.assertTrue(bucket['bucket_quota']['max_size_kb'] == size)

    async def test_bucket(self):
        bucket_name = f"bucket-{uuid.uuid4()}"
        await create_bucket(name=bucket_name, owner=self.user1)
        bucket = await self.rgw.get_bucket(bucket=bucket_name)
        await self.rgw.link_bucket(bucket=bucket_name, bucket_id=bucket['id'],
                                   uid=self.user1)
        await self.rgw.get_bucket(uid=self.user1, bucket=bucket_name)
        await self.rgw.get_policy(bucket=bucket_name)
        await self.rgw.remove_bucket(bucket=bucket_name, purge_objects=True)
        self.assertTrue(True)

    async def test_get_usage(self):
        summary = await self.rgw.get_usage(show_summary=True)
        self.assertTrue('summary' in summary)

    async def test_subuser(self):
        await self.rgw.create_subuser(uid=self.user2,
                                      subuser='swift',
                                      key_type='swift',
                                      secret_key=self.secret)
        await self.rgw.modify_subuser(uid=self.user2,
                                      subuser='swift',
                                      access='write')
        subuser = await self.rgw.get_user(uid=self.user2)
        await self.rgw.remove_subuser(uid=self.user2, subuser='swift')
        for subs in subuser['subusers']:
            if subs['id'] == '%s:%s' % (self.user2, 'swift'):
                self.assertTrue(subs['permissions'] == 'write')

    async def test_s3_keys(self):
        access = aiorgwadmin.RGWAdmin.gen_secret_key(size=20)
        secret = aiorgwadmin.RGWAdmin.gen_secret_key(size=40)
        keys = await self.rgw.create_key(uid=self.user1,
                                         access_key=access,
                                         secret_key=secret)
        for key in keys:
            if key['access_key'] == access:
                self.assertTrue(key['secret_key'] == secret)

    def test_parse_rados_datestring(self):
        aiorgwadmin.RGWAdmin.parse_rados_datestring(u'2016-06-27T16:06:39.163Z')


if __name__ == '__main__':
    unittest.main()



import logging
import sys
from collections import OrderedDict

from typing import Any, Final, List

from .utils import random_password
from .rgw import RGWAdmin
from .exceptions import NoSuchKey

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

log: Final = logging.getLogger(__name__)


class AttributeMixin(object):
    attrs: List[str] = []

    def __str__(self) -> str:
        try:
            from qav.listpack import ListPack
        except ImportError:
            return str(self.to_tuples())
        return str(ListPack(self.to_tuples()))

    def to_tuples(self) -> list:
        return [(x, getattr(self, x)) for x in self.attrs]

    def to_dict(self) -> dict:
        return dict(self.to_tuples())


class RGWCap(AttributeMixin):
    attrs = ['type', 'perm']

    def __init__(self, type: str, perm: str) -> None:
        self.type = type
        self.perm = perm

    def __repr__(self) -> str:
        return str('%s %s' % (self.__class__.__name__, self.type))


class RGWKey(AttributeMixin):
    attrs = ['user', 'access_key', 'secret_key']

    def __init__(self, user: str, access_key: str, secret_key: str) -> None:
        self.user = user
        self.access_key = access_key
        self.secret_key = secret_key

    def __repr__(self) -> str:
        return str('%s %s' % (self.__class__.__name__, self.user))

    @classmethod
    def generate(cls, user: str, access_size: int = 20, secret_size: int = 40) -> Self:
        access_key = random_password(size=access_size)
        secret_key = random_password(size=secret_size)
        return cls(user, access_key, secret_key)


class RGWSubuser(AttributeMixin):
    attrs = ['id', 'permissions']
    permission_list = ['read', 'write', 'read-write', 'full-control']

    def __init__(self, id: str, permissions: str) -> None:
        self.id = id
        self.permissions = permissions

    def __repr__(self) -> str:
        return str('%s %s' % (self.__class__.__name__, self.id))

    @property
    def permissions(self) -> str:
        return self._permissions

    @permissions.setter
    def permissions(self, value: str) -> None:
        if value in self.permission_list:
            self._permissions = value
        elif value == '<none>':
            self._permissions = None
        else:
            raise AttributeError


class RGWSwiftKey(AttributeMixin):
    attrs = ['user', 'secret_key']

    def __init__(self, user: str, secret_key: str) -> None:
        self.user = user
        self.secret_key = secret_key

    def __repr__(self) -> str:
        return str('%s %s' % (self.__class__.__name__, self.user))

    @classmethod
    def generate(cls, user: str, secret_size: int = 40) -> Self:
        secret_key = random_password(size=secret_size)
        return cls(user, secret_key)


class RGWQuota(AttributeMixin):
    attrs = ['enabled', 'max_objects', 'max_size_kb']
    defaults = OrderedDict([('enabled', False), ('max_objects', -1), ('max_size_kb', -1)])

    def __init__(self, **kwargs: Any) -> None:
        for attr in self.attrs:
            setattr(self, attr, kwargs[attr])

    def __repr__(self) -> str:
        return '%s %s/%s' % (self.enabled, self.max_objects, self.string_size())

    def string_size(self) -> int | str:
        weights = ['KB', 'MB', 'GB', 'TB', 'PB']
        size = int(self.max_size_kb)
        weight = 0
        if size < 0:
            return size
        while True:
            if size < 1024 or weight == len(weights):
                return str('%s%s' % (size, weights[weight]))
            else:
                size = size / 1024
                weight += 1

    @classmethod
    def default(cls) -> Self:
        return cls(**cls.defaults)

    @property
    def size(self) -> int:
        return int(self.max_size_kb) * 1000

    @size.setter
    def size(self, new_size: int) -> None:
        self.max_size_kb = int(new_size) / 1000


class RGWUser(AttributeMixin):
    attrs = ['user_id', 'display_name', 'email', 'caps', 'keys',
             'max_buckets', 'suspended', 'swift_keys', 'subusers',
             'placement_tags', 'bucket_quota', 'user_quota',
             'default_placement', 'op_mask', 'temp_url_keys']
    modify_attrs_mask = ['placement_tags', 'default_placement', 'op_mask', 'temp_url_keys']
    sub_attrs = OrderedDict([('caps', RGWCap),
                             ('keys', RGWKey),
                             ('swift_keys', RGWSwiftKey),
                             ('subusers', RGWSubuser),
                             ('bucket_quota', RGWQuota),
                             ('user_quota', RGWQuota)])

    """Representation of a RadosGW User"""

    def __init__(self, **kwargs: Any) -> None:
        self.sensitive_attrs = [('keys', 'secret_key')]
        for attr in self.attrs:
            setattr(self, attr, kwargs[attr])
        if kwargs.get('sensitive_attrs'):
            self.sensitive_attrs.append(kwargs.get('sensitive_attrs'))

    def __repr__(self) -> str:
        return str('%s %s' % (self.__class__.__name__, self.user_id))

    @classmethod
    async def create(cls, user_id: str, display_name: str, **kwargs: Any) -> Self | None:
        rgw = RGWAdmin.get_connection()
        log.debug('Creating user %s' % user_id)
        await rgw.create_user(uid=user_id, display_name=display_name, **kwargs)
        return await cls.fetch(user_id)

    async def diff(self) -> dict:
        rgw = RGWAdmin.get_connection()
        d = {}
        current = self.to_dict()
        try:
            existing = await rgw.get_metadata('user', self.user_id)['data']
        except NoSuchKey:
            logging.info('Unable to fetch the user %s.' % self.user_id)
            return current
        for k in current.keys():
            if k in existing.keys():
                if current[k] != existing[k]:
                    d[k] = (existing[k], current[k])
        return d

    async def exists(self) -> bool:
        """Return True if the user exists.  False otherwise."""
        rgw = RGWAdmin.get_connection()
        try:
            await rgw.get_metadata('user', self.user_id)
            return True
        except NoSuchKey:
            return False

    async def save(self) -> None:
        rgw = RGWAdmin.get_connection()

        if not await self.exists():
            log.debug('User does not exist. Creating %s' % self.user_id)
            await rgw.create_user(uid=self.user_id, display_name=self.display_name)

        d = self._modify_dict()
        log.debug('Modify existing user %s %s' % (self.user_id, d))
        await rgw.modify_user(**d)
        await rgw.set_user_quota(self.user_id, 'user', **self.user_quota.to_dict())
        await rgw.set_user_quota(self.user_id, 'bucket', **self.bucket_quota.to_dict())

    async def delete(self) -> Any:
        rgw = RGWAdmin.get_connection()
        try:
            await rgw.get_metadata('user', self.user_id)
        except NoSuchKey:
            logging.error('User %s does not exist yet aborting.' % self.user_id)
            return False
        logging.info('Deleting user %s from %s' % (self.user_id, rgw.get_base_url()))
        return await rgw.remove_user(uid=self.user_id)

    @classmethod
    async def list(cls) -> list:
        rgw = RGWAdmin.get_connection()
        return [await cls.fetch(x) for x in await rgw.get_users()]

    @classmethod
    async def fetch(cls, user: str) -> Self | None:
        log.debug('Fetching user %s' % user)
        rgw = RGWAdmin.get_connection()
        try:
            j = await rgw.get_metadata('user', user)
        except NoSuchKey:
            return None
        else:
            return cls._parse_user(j['data'])

    @classmethod
    def _parse_user(cls, rgw_user: Any) -> Self | None:
        # we expect to be passed a dict
        if type(rgw_user) is not dict:
            log.warning('rgw_user is not a dict instance')
            return None
        # check to make sure we have all the correct keys
        if not set([str(x) for x in cls.attrs]) <= \
                set(rgw_user.keys()):
            return None
        log.debug('Parsing RGWUser %s' % rgw_user)
        for subattr in cls.sub_attrs.keys():
            log.debug('Loading attribute %s with class %s' %
                      (subattr, cls.sub_attrs[subattr].__name__))
            if type(rgw_user[subattr]) is list:
                obj = [cls.sub_attrs[subattr](**x) for x in rgw_user[subattr]]
            elif type(rgw_user[subattr]) is dict:
                obj = cls.sub_attrs[subattr](**rgw_user[subattr])
            else:
                obj = rgw_user[subattr]
            rgw_user[subattr] = obj
        return cls(**rgw_user)

    def _scrubbed_dict(self) -> dict:
        '''Return a dict representation of the object with sensitve attrs
           filtered.
        '''
        scrubbed = self.to_dict()
        censor = '******'
        for k, v in self.sensitive_attrs:
            if type(scrubbed[k]) is list:
                for o in scrubbed[k]:
                    o[v] = censor
            else:
                scrubbed[k] = censor
        return scrubbed

    def _modify_dict(self) -> dict:
        '''Return flat dict representation of the object'''
        d = {}
        for attr in self.attrs:
            if attr not in self.modify_attrs_mask+list(self.sub_attrs.keys()):
                d[attr] = getattr(self, attr)
        d['uid'] = d.pop('user_id')
        d['generate_key'] = False
        return d

    def to_dict(self) -> dict:
        '''Return the dict representation of the object'''
        d = {}
        for attr in self.attrs:
            if attr in self.sub_attrs.keys():
                if type(getattr(self, attr)) is list:
                    d[attr] = []
                    for o in getattr(self, attr):
                        d[attr].append(o.to_dict())
                else:
                    d[attr] = getattr(self, attr).to_dict()
            else:
                d[attr] = getattr(self, attr)
        return d



import string
import random
from collections.abc import Sequence


def random_password(size: int = 40, chars: Sequence | None = None) -> str:
    '''Return a random password string.

    The default length is 32 characters.  Different character classes can be
    passed in, but the default draws from ASCII uppercase, lowercase, and
    digits.

    Uses the SystemRamdom to ensure you get a trustworty entropy pool.
    '''
    if chars is None:
        chars = string.ascii_uppercase + string.ascii_lowercase + string.digits
    random_chars = random.SystemRandom().choice
    return ''.join(random_chars(chars) for _ in range(size))



"""
aiorgwadmin.exceptions
~~~~~~~~~~~~~~~~~~~

This module contains the aiorgwadmin exceptions.

"""
from typing import Any


class RGWAdminException(Exception):
    """There was a unlabeled exception that was raised during your request"""
    def __init__(self, code: Any, raw: Any = None) -> None:
        self.code = code
        self.raw = raw


class AccessDenied(RGWAdminException):
    """Access was denied for the request."""


class UserExists(RGWAdminException):
    """Attempt to create existing user."""


class InvalidAccessKey(RGWAdminException):
    """Invalid access key specified."""


class InvalidArgument(RGWAdminException):
    """Invalid argument specified."""


class InvalidKeyType(RGWAdminException):
    """Invalid key type specified."""


class InvalidSecretKey(RGWAdminException):
    """Invalid secret key specified."""


class KeyExists(RGWAdminException):
    """Provided access key exists and belongs to another user."""


class EmailExists(RGWAdminException):
    """Provided email address exists."""


class SubuserExists(RGWAdminException):
    """Specified subuser exists."""


class InvalidAccess(RGWAdminException):
    """Invalid subuser access specified."""


class IndexRepairFailed(RGWAdminException):
    """Bucket index repair failed."""


class BucketNotEmpty(RGWAdminException):
    """Attempted to delete non-empty bucket."""


class ObjectRemovalFailed(RGWAdminException):
    """Unable to remove objects."""


class BucketUnlinkFailed(RGWAdminException):
    """Unable to unlink bucket from specified user."""


class BucketLinkFailed(RGWAdminException):
    """Unable to link bucket to specified user."""


class NoSuchObject(RGWAdminException):
    """Specified object does not exist."""


class IncompleteBody(RGWAdminException):
    """Either bucket was not specified for a bucket policy request or bucket
       and object were not specified for an object policy request."""


class InvalidCap(RGWAdminException):
    """Attempt to grant invalid admin capability."""


class NoSuchCap(RGWAdminException):
    """User does not possess specified capability."""


class InternalError(RGWAdminException):
    """Internal server error."""


class NoSuchUser(RGWAdminException):
    """User does not exist."""


class NoSuchBucket(RGWAdminException):
    """Bucket does not exist."""


class NoSuchKey(RGWAdminException):
    """No such access key."""


class ServerDown(RGWAdminException):
    """The backing server is not available."""


class InvalidQuotaType(RGWAdminException):
    """You must specify either a 'user' or 'bucket' quota type"""


class BucketAlreadyExists(RGWAdminException):
    """The bucket already exists"""



#!/usr/bin/env python

import logging
import unittest
import uuid

import aiorgwadmin
from aiorgwadmin.user import RGWUser
from . import get_environment_creds

logging.basicConfig(level=logging.WARNING)


class RGWUserTest(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.rgw = aiorgwadmin.RGWAdmin(**get_environment_creds())
        aiorgwadmin.RGWAdmin.set_connection(self.rgw)

        self.user = f"user-{uuid.uuid4()}"

    async def asyncTearDown(self):
        users = await self.rgw.get_users()

        if self.user in users:
            await self.rgw.remove_user(uid=self.user, purge_data=True)

    async def test_create_user(self):
        display_name = f"Test Create User"
        u = await RGWUser.create(user_id=self.user, display_name=display_name)
        self.assertTrue(u.user_id == self.user and
                        u.display_name == display_name)
        await u.delete()

    async def test_user_exists(self):
        display_name = "Test User Exists"
        u = await RGWUser.create(user_id=self.user, display_name=display_name)
        self.assertTrue(await u.exists())
        await u.delete()
        self.assertFalse(await u.exists())
        await u.save()
        self.assertTrue(await u.exists())

    async def test_set_quota(self):
        display_name = "Test Set Quota"
        u = await RGWUser.create(user_id=self.user, display_name=display_name)
        u.user_quota.size = 1024000
        await u.save()
        nu = await RGWUser.fetch(u.user_id)
        self.assertTrue(u.user_quota.size == nu.user_quota.size)
        await nu.delete()


if __name__ == '__main__':
    unittest.main()



__title__ = 'aiorgwadmin'
__version__ = "1.1.0"
__author__ = 'Derek Yarnell, Mikle Green'
__license__ = 'LGPL v2.1'

from .rgw import RGWAdmin
from .user import RGWUser, RGWQuota, RGWSwiftKey, RGWSubuser, RGWKey, RGWCap



#!/usr/bin/env python

from os import path
import ast
import re

try:
    from setuptools import setup
    extra = dict(include_package_data=True)
except ImportError:
    from distutils.core import setup
    extra = {}

BASE_DIR = path.abspath(path.dirname(__file__))
with open(path.join(BASE_DIR, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

install_requires = [
    "aiohttp",
    "requests",
    "requests-aws",
]

_version_re = re.compile(r'__version__\s+=\s+(.*)')
with open('aiorgwadmin/__init__.py', encoding='utf-8') as f:
    version = str(ast.literal_eval(_version_re.search(f.read()).group(1)))


setup(
    name="aiorgwadmin",
    packages=["aiorgwadmin"],
    package_data={"aiorgwadmin": ["py.typed"]},
    zip_safe=False,
    version=version,
    install_requires=install_requires,
    author="Derek Yarnell <derek@umiacs.umd.edu>, Mikle Green",
    url="https://github.com/mikle-green/aiorgwadmin",
    license="LGPL v2.1",
    description="Python Rados Gateway Admin API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["ceph", "radosgw", "admin api", "async"],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
    ],
    python_requires=">=3.10",
    **extra
)



#!/usr/bin/env python

import json
import logging
import unittest
import uuid
from urllib.parse import quote

import aiorgwadmin
from . import create_bucket, get_environment_creds

logging.basicConfig(level=logging.WARNING)


class MetadataTest(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.rgw = aiorgwadmin.RGWAdmin(**get_environment_creds())
        aiorgwadmin.RGWAdmin.set_connection(self.rgw)

        self.user = f"user-{uuid.uuid4()}"
        await self.rgw.create_user(uid=self.user, display_name=f"Unit Test {self.user}")

    async def asyncTearDown(self):
        await self.rgw.remove_user(uid=self.user, purge_data=True)

    async def test_get_metadata(self):
        bucket_name = f"bucket-{uuid.uuid4()}"
        self.assertTrue(bucket_name not in await self.rgw.get_metadata('bucket'))

        await create_bucket(bucket_name, owner=self.user)
        self.assertTrue(bucket_name in await self.rgw.get_metadata('bucket'))

    async def test_put_metadata(self):
        bucket_name = f"bucket-{uuid.uuid4()}"
        await create_bucket(bucket_name, owner=self.user)

        ret_json = await self.rgw.get_metadata('bucket', key=bucket_name)
        self.assertEqual(ret_json['data']['bucket']['name'], bucket_name)
        json_str = json.dumps(ret_json)

        await self.rgw.put_metadata('bucket', key=bucket_name, json_string=json_str)

    async def test_metadata_lock_unlock(self):
        bucket_name = f"bucket-{uuid.uuid4()}"
        await create_bucket(bucket_name, owner=self.user)

        await self.rgw.lock_metadata('bucket', key=bucket_name, lock_id='abc',
                                     length=5)
        await self.rgw.unlock_metadata('bucket', key=bucket_name, lock_id='abc')

    async def test_invalid_metadata_unlock(self):
        with self.assertRaises(aiorgwadmin.exceptions.NoSuchKey):
            key = f"bucket-{uuid.uuid4()}"
            await self.rgw.unlock_metadata('bucket', key=key, lock_id='abc')

    async def test_metadata_type_valid(self):
        with self.assertRaises(Exception):
            await self.rgw.get_metadata('bucketttt')

    async def test_get_bucket_instances(self):
        bucket_name = f"bucket-{uuid.uuid4()}"
        await create_bucket(bucket_name, owner=self.user)

        instances = await self.rgw.get_bucket_instances()
        bucket = await self.rgw.get_bucket(bucket_name)
        expected_instance = '%s:%s' % (bucket_name, bucket['id'])
        self.assertTrue(expected_instance in instances)

    def test_metadata_marker(self):
        self.assertEqual('default.345%20-5', quote('default.345 -5'))


if __name__ == '__main__':
    unittest.main()
